/* player.c
 *
 * Copyright 2002-2003 Vesa Halttunen
 *
 * This file is part of Tutka.
 *
 * Tutka is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Tutka is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tutka; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <gtk/gtk.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <glib.h>
#include <sched.h>
#include <linux/rtc.h>
#include "song.h"
#include "midi.h"
#include "player.h"
#include "editor.h"

#define RTC_FREQ 512

/* Refreshes playseq from section and block from position */
static void player_refresh_playseq_and_block(struct player *player)
{
  struct song *song;

  if (player == NULL) {
    fprintf(stderr, "player_refresh_playseq_and_block() called with null player\n");
    return;
  }

  song = player->song;

  if (player->section >= song->numsections)
    player->section = 0;
  player->playseq = song->sections[player->section];

  if (player->position >= song->playseqs[player->playseq]->length) {
    player->position = 0;
    player_refresh_playseq_and_block(player);
  } else
    player->block = song->playseqs[player->playseq]->blocknumbers[player->position];
}

/* Advances in section and jumps to the beginning if necessary */
static void player_next_section(struct player *player)
{
  if (player == NULL) {
    fprintf(stderr, "player_next_section() called with null player\n");
    return;
  }

  player->section++;
  if (player->section >= player->song->numsections)
    player->section = 0;
}

/* Advances in playing sequence and jumps to next section if necessary */
static void player_next_playseq(struct player *player)
{
  if (player == NULL) {
    fprintf(stderr, "player_next_playseq() called with null player\n");
    return;
  }

  player->position++;
  if (player->position >= player->song->playseqs[player->playseq]->length) {
    player->position = 0;
    player_next_section(player);
  }
}

/* Plays a note using given instrument on a given track */
void player_play_note(struct player *player, unsigned int instrument, unsigned char note, unsigned char volume, unsigned char track)
{
  struct song *song;

  if (player == NULL) {
    fprintf(stderr, "player_play_note() called with null player\n");
    return;
  }

  song = player->song;

  /* Stop currently playing note */
  if (player->trackstatus[track * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE] != -1) {
    midi_note_off(player->midi, player->trackstatus[track * TRACK_STATUS_NUMVALUES + TRACK_STATUS_MIDI_CHANNEL], player->trackstatus[track * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE], 127);
    player->trackstatus[track * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE] = -1;
  }

  /* Don't play a note if the instrument does not exist */
  if (instrument < song->numinstruments && song->instruments[instrument] != NULL) {
    /* Apply note transpose */
    note += song->instruments[instrument]->transpose;

    /* Set new track status (channel, note, volume, time) */
    player->trackstatus[track * TRACK_STATUS_NUMVALUES + TRACK_STATUS_MIDI_CHANNEL] = song->instruments[instrument]->midichannel;
    player->trackstatus[track * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE] = note;
    player->trackstatus[track * TRACK_STATUS_NUMVALUES + TRACK_STATUS_VOLUME] = song->instruments[instrument]->defaultvelocity * volume / 127 * (song->trackvolumes[track] & 127) / 127 * ((255 - song->trackvolumes[track]) & 128) / 128 * song->mastervolume / 127;
    if (song->instruments[instrument]->hold > 0)
      player->trackstatus[track * TRACK_STATUS_NUMVALUES + TRACK_STATUS_HOLD] = song->instruments[instrument]->hold;
    else
      player->trackstatus[track * TRACK_STATUS_NUMVALUES + TRACK_STATUS_HOLD] = -1;

    /* Make sure the volume isn't too large */
    if (player->trackstatus[track * TRACK_STATUS_NUMVALUES + TRACK_STATUS_VOLUME] < 0)
      player->trackstatus[track * TRACK_STATUS_NUMVALUES + TRACK_STATUS_VOLUME] = 127;

    /* Play note */
    midi_note_on(player->midi, player->trackstatus[track * TRACK_STATUS_NUMVALUES + TRACK_STATUS_MIDI_CHANNEL], player->trackstatus[track * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE], player->trackstatus[track * TRACK_STATUS_NUMVALUES + TRACK_STATUS_VOLUME]);
  }
}

/* Stops notes playing at the moment */
void player_stop_notes(struct player *player)
{
  int i;

  if (player == NULL) {
    fprintf(stderr, "player_stop_notes() called with null player\n");
    return;
  }

  for (i = 0; i < player->song->maxtracks; i++) {
    if (player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE] != -1) {
      midi_note_off(player->midi, player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_MIDI_CHANNEL], player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE], 127);
      player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_MIDI_CHANNEL] = -1;
      player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE] = -1;
      player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_VOLUME] = -1;
      player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_HOLD] = -1;
    }
  }
}

/* Stops all notes playing at the moment */
void player_stop_all_notes(struct player *player)
{
  int i, j;

  if (player == NULL) {
    fprintf(stderr, "player_stop_all_notes() called with null player\n");
    return;
  }

  for (i = 0; i < 16; i++)
    for (j = 0; j < 128; j++)
      midi_note_off(player->midi, i, j, 127);
}

/* Resets the pitch wheel on all channels */
void player_reset_pitch(struct player *player)
{
  int i;

  if (player == NULL) {
    fprintf(stderr, "player_reset_pitch() called with null player\n");
    return;
  }

  for (i = 0; i < 16; i++)
    midi_pitch_wheel(player->midi, i, 64);
}

/* Player thread */
static void player_thread(struct player *player)
{
  struct timespec req, rem;
  struct timeval next, now;

  if (player == NULL) {
    fprintf(stderr, "player_thread() called with null player\n");
    return;
  }

  player->tick = 0;

  if (player->sched == SCHED_RTC) {
    if (ioctl(player->rtc, RTC_UIE_OFF, 0) == -1 || ioctl(player->rtc, RTC_IRQP_SET, RTC_FREQ) == -1 || ioctl(player->rtc, RTC_PIE_ON, 0) == -1)
      player->sched = SCHED_NANOSLEEP;
  }

  next.tv_sec = player->playingstarted.tv_sec;
  next.tv_usec = player->playingstarted.tv_usec;

  while (1) {
    struct timeval tod;
    int i, j, changetempo = 0;
    struct song *song;
    struct block *b;
    unsigned char postcommand = 0, postvalue = 0;

    /* Calculate time of next tick (tick every 1000000/((BPM/60)*24) usecs) */
    next.tv_sec += (25 / player->song->tempo) / 10;
    next.tv_usec += ((2500000 / player->song->tempo) % 1000000);
    while (next.tv_usec > 1000000) {
      next.tv_sec++;
      next.tv_usec -= 1000000;
    }

    gettimeofday(&now, NULL);

    /* Select scheduling type */
    if (player->sched == SCHED_NANOSLEEP) {
      /* Nanosleep: calculate difference between now and the next tick */
      req.tv_sec = next.tv_sec - now.tv_sec;
      req.tv_nsec = (next.tv_usec - now.tv_usec) * 1000;
      while (req.tv_nsec < 0) {
	req.tv_sec--;
	req.tv_nsec += 1000000000;
      }

      /* Sleep until the next tick if necessary */
      if (req.tv_sec >= 0)
	while (nanosleep(&req, &rem) == -1);
    } else {
      /* Rtc: read RTC interrupts until it's time to play again */
      while ((next.tv_sec - now.tv_sec) > 0 || ((next.tv_sec - now.tv_sec) == 0 && (next.tv_usec - now.tv_usec) > 1000000 / RTC_FREQ)) {
	unsigned long l;
	read(player->rtc, &l, sizeof(unsigned long));
	gettimeofday(&now, NULL);
      }
      while ((next.tv_sec - now.tv_sec) > 0 || ((next.tv_sec - now.tv_sec) == 0 && (next.tv_usec - now.tv_usec) > 0)) {
	gettimeofday(&now, NULL);
      }
    }

    /* Handle this tick */
    song = player->song;
    b = song->blocks[player->block];

    /* Send MIDI sync if requested */
    if (song->sendsync)
      midi_clock(player->midi);

    /* Pass 1: Stop notes if there are new notes about to be played */
    for (i = 0; i < b->tracks; i++) {
      int delay = 0;
      unsigned char note = b->notes[(b->tracks * player->line + i) * 2];

      if (note != 0) {
	for (j = 0; j < b->commandpages; j++) {
	  unsigned char command = b->commands[j * b->tracks * b->length * 2 + (b->tracks * player->line + i) * 2];
	  unsigned char value = b->commands[j * b->tracks * b->length * 2 + (b->tracks * player->line + i) * 2 + 1];

	  if (command != 0 || value != 0) {
	    /* Check for previous command if any */
	    if (command == COMMAND_PREVIOUS_COMMAND_VALUE) {
	      if (value != 0)
		command = player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_PREVIOUS_COMMAND];
	    } else
	      player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_PREVIOUS_COMMAND] = command;

	    switch (command) {
	    case COMMAND_RETRIGGER:
	      delay = -value;
	      break;
	    case COMMAND_DELAY:
	      delay = value;
	      break;
	    }
	  }
	}

	/* Stop currently playing note */
	if ((delay >= 0 && player->tick == delay) || (delay < 0 && player->tick % (-delay) == 0)) {
	  if (player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE] != -1) {
	    midi_note_off(player->midi, player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_MIDI_CHANNEL], player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE], 127);
	    player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE] = -1;
	  }
	}
      }
    }

    /* Pass 2: Play notes scheduled to be played */
    for (i = 0; i < b->tracks; i++) {
      unsigned int volume = 127;
      int delay = 0, hold = -1;
      unsigned char note = b->notes[(b->tracks * player->line + i) * 2];
      unsigned char instrument = b->notes[(b->tracks * player->line + i) * 2 + 1];

      /* Handle commands on all command pages */
      for (j = 0; j < b->commandpages; j++) {
	unsigned char command = b->commands[j * b->tracks * b->length * 2 + (b->tracks * player->line + i) * 2];
	unsigned char value = b->commands[j * b->tracks * b->length * 2 + (b->tracks * player->line + i) * 2 + 1];
	int midichannel = -1;

	if (command != 0 || value != 0) {
	  /* Check which MIDI channel the command will affect */
	  if (instrument != 0)
	    /* Instrument number defines MIDI channel */
	    midichannel = song->instruments[instrument - 1]->midichannel;
	  else
	    /* Note playing defines MIDI channel */
	    midichannel = player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_MIDI_CHANNEL];

	  /* Check for previous command if any */
	  if (command == COMMAND_PREVIOUS_COMMAND_VALUE) {
	    if (value != 0)
	      command = player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_PREVIOUS_COMMAND];
	  } else
	    player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_PREVIOUS_COMMAND] = command;

	  switch (command) {
	  case COMMAND_PITCH_WHEEL:
	    /* Pitch wheel can be set if the MIDI channel is known */
	    if (midichannel != -1) {
	      if (value < 0x80) {
		if (player->tick == 0) {
		  midi_pitch_wheel(player->midi, midichannel, value);
		  player->midicontrollervalues[midichannel * VALUES + VALUES_PITCH_WHEEL] = value;
		}
	      } else {
		if (player->tick < song->ticksperline - 1) {
		  int delta = (value - 0x80 - player->midicontrollervalues[midichannel * VALUES + VALUES_PITCH_WHEEL]) / ((int)song->ticksperline - 1);

		  midi_pitch_wheel(player->midi, midichannel, player->midicontrollervalues[midichannel * VALUES + VALUES_PITCH_WHEEL] + player->tick * delta);
		} else {
		  midi_pitch_wheel(player->midi, midichannel, value - 0x80);
		  player->midicontrollervalues[midichannel * VALUES + VALUES_PITCH_WHEEL] = value - 0x80;
		}
	      }
	    }
	    break;
	  case COMMAND_END_BLOCK:
	    /* Only on last tick */
	    if (player->tick == song->ticksperline - 1) {
	      postcommand = COMMAND_END_BLOCK;
	      postvalue = value;
	    }
	    break;
	  case COMMAND_PLAYSEQ_POSITION:
	    /* Only on last tick */
	    if (player->tick == song->ticksperline - 1) {
	      postcommand = COMMAND_PLAYSEQ_POSITION;
	      postvalue = value;
	    }
	    break;
	  case COMMAND_SEND_MESSAGE:
	    /* Only on first tick */
	    if (player->tick == 0 && value < song->nummessages)
	      midi_write(player->midi, song->messages[value]->data, song->messages[value]->length);
	    break;
	  case COMMAND_HOLD:
	    hold = value;
	    break;
	  case COMMAND_RETRIGGER:
	    delay = -value;
	    break;
	  case COMMAND_DELAY:
	    delay = value;
	    break;
	  case COMMAND_VELOCITY:
	    if (note != 0) {
	      volume = value;
	      if (midichannel != -1)
		player->midicontrollervalues[midichannel * VALUES + VALUES_AFTERTOUCH] = value;
	    } else {
	      /* Note playing defines MIDI channel */
	      midichannel = player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_MIDI_CHANNEL];

	      if (midichannel != -1) {
		if (value < 0x80) {
		  if (player->tick == 0) {
		    if (value > 0) {
		      midi_aftertouch(player->midi, midichannel, player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE], value);
		      player->midicontrollervalues[midichannel * VALUES + VALUES_AFTERTOUCH] = value;
		    } else {
		      midi_note_off(player->midi, midichannel, player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE], 127);
		      player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE] = -1;
		    }
		  }
		} else {
		  if (player->tick < song->ticksperline - 1) {
		    int delta = (value - 0x80 - player->midicontrollervalues[midichannel * VALUES + VALUES_AFTERTOUCH]) / ((int)song->ticksperline - 1);

		    midi_aftertouch(player->midi, midichannel, player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE], player->midicontrollervalues[midichannel * VALUES + VALUES_AFTERTOUCH] + player->tick * delta);
		  } else {
		    midi_aftertouch(player->midi, midichannel, player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE], value - 0x80);
		    player->midicontrollervalues[midichannel * VALUES + VALUES_AFTERTOUCH] = value - 0x80;
		  }
		}
	      }
	    }
	    break;
	  case COMMAND_CHANNEL_PRESSURE:
	    /* Channel pressure can be set if the MIDI channel is known */
	    if (midichannel != -1) {
	      if (value < 0x80) {
		if (player->tick == 0) {
		  midi_channel_pressure(player->midi, midichannel, value);
		  player->midicontrollervalues[midichannel * VALUES + VALUES_CHANNEL_PRESSURE] = value;
		}
	      } else {
		if (player->tick < song->ticksperline - 1) {
		  int delta = (value - 0x80 - player->midicontrollervalues[midichannel * VALUES + VALUES_CHANNEL_PRESSURE]) / ((int)song->ticksperline - 1);

		  midi_channel_pressure(player->midi, midichannel, player->midicontrollervalues[midichannel * VALUES + VALUES_CHANNEL_PRESSURE] + player->tick * delta);
		} else {
		  midi_channel_pressure(player->midi, midichannel, value - 0x80);
		  player->midicontrollervalues[midichannel * VALUES + VALUES_CHANNEL_PRESSURE] = value - 0x80;
		}
	      }
	    }
	    break;
	  case COMMAND_TPL:
	    if (value == 0) {
	      /* Only on last tick */
	      if (player->tick == song->ticksperline - 1)
		postcommand = COMMAND_TPL;
	    } else {
	      song_set_tpl(song, value);
	      changetempo = 1;
	    }
	    break;
	  case COMMAND_TEMPO:
	    if (value == 0) {
	      /* Only on last tick */
	      if (player->tick == song->ticksperline - 1)
		postcommand = COMMAND_TEMPO;
	    } else {
	      song_set_tempo(song, value);
	      changetempo = 1;
	    }
	    break;
	  }

	  /* Handle MIDI controllers */
	  if (command >= COMMAND_MIDI_CONTROLLERS) {
	    /* MIDI controllers can be set if the MIDI channel is known */
	    if (midichannel != -1) {
	      if (value < 0x80) {
		if (player->tick == 0) {
		  midi_controller(player->midi, midichannel, command - COMMAND_MIDI_CONTROLLERS, value);
		  player->midicontrollervalues[midichannel * VALUES + command - COMMAND_MIDI_CONTROLLERS] = value;
		}
	      } else {
		if (player->tick < song->ticksperline - 1) {
		  int delta = (value - 0x80 - player->midicontrollervalues[midichannel * VALUES + command - COMMAND_MIDI_CONTROLLERS]) / ((int)song->ticksperline - 1);

		  midi_controller(player->midi, midichannel, command - COMMAND_MIDI_CONTROLLERS, player->midicontrollervalues[midichannel * VALUES + command - COMMAND_MIDI_CONTROLLERS] + player->tick * delta);
		} else {
		  midi_controller(player->midi, midichannel, command - COMMAND_MIDI_CONTROLLERS, value - 0x80);
		  player->midicontrollervalues[midichannel * VALUES + command - COMMAND_MIDI_CONTROLLERS] = value - 0x80;
		}
	      }
	    }
	  }
	}
      }

      /* Is there a note to play? */
      if (note != 0 && ((delay >= 0 && player->tick == delay) || (delay < 0 && player->tick % (-delay) == 0))) {
	note--;

	player_play_note(player, instrument - 1, note, volume, i);

	/* If no hold value was defined use the instrument's hold value */
	if (hold == -1 && instrument <= song->numinstruments && song->instruments[instrument - 1] != NULL)
	  hold = song->instruments[b->notes[(b->tracks * player->line + i) * 2 + 1] - 1]->hold;

	if (hold == 0)
	  player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_HOLD] = -1;
	else
	  player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_HOLD] = hold;
      }
      
      /* First tick, no note but instrument defined? */
      if (player->tick == 0 && note == 0 && instrument > 0 && player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_HOLD] >= 0)
        player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_HOLD] += song->instruments[instrument - 1]->hold;
    }

    /* Decrement hold times of notes and stop notes that should be stopped */
    for (i = 0; i < song->maxtracks; i++) {
      if (player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_HOLD] >= 0) {
	player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_HOLD]--;
	if (player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_HOLD] < 0) {
	  midi_note_off(player->midi, player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_MIDI_CHANNEL], player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE], 127);
	  player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE] = -1;
	}
      }
    }

    /* Next tick */
    player->tick++;
    player->tick %= song->ticksperline;

    /* Advance and handle post commands if ticksperline ticks have passed */
    if (player->tick == 0) {
      int changeblock = 0;

      player->line++;

      switch (postcommand) {
      case COMMAND_END_BLOCK:
	player->line = postvalue;
	if (player->mode == MODE_PLAY_SONG) {
	  player_next_playseq(player);
	  changeblock = 1;
	}
	break;
      case COMMAND_PLAYSEQ_POSITION:
	player->line = 0;
	player->position = postvalue;
	if (player->position >= song->playseqs[player->playseq]->length) {
	  player->position = 0;
	  player_next_section(player);
	}
	changeblock = 1;
	break;
      case COMMAND_TEMPO:
	/* COMMAND_TPL and COMMAND_TEMPO can only mean "stop" as stop cmds */
	g_mutex_lock(player->mutex);
	player->killthread = 1;
	g_mutex_unlock(player->mutex);
	return;
	break;
      default:
	/* Advance in block */
	if (player->line >= song->blocks[player->block]->length) {
	  player->line = 0;
	  if (player->mode == MODE_PLAY_SONG) {
	    player_next_playseq(player);
	    changeblock = 1;
	  }
	}
	break;
      }

      if (changeblock == 1)
	player_refresh_playseq_and_block(player);

      gettimeofday(&tod, NULL);

      /* Tell the editor about the new position if an editor exists */
      if (player->editor != NULL) {
	editor_set_section(player->editor, player->section);
	editor_set_playseq(player->editor, player->playseq);
	editor_set_position(player->editor, player->position);
	editor_set_block(player->editor, player->block);
	editor_set_line(player->editor, player->line);
	editor_set_time(player->editor, (int)(player->playedsofar.tv_sec * 1000 + player->playedsofar.tv_usec / 1000 + (tod.tv_sec * 1000 + tod.tv_usec / 1000) - (player->playingstarted.tv_sec * 1000 + player->playingstarted.tv_usec / 1000)) / 1000);
	if (changetempo)
	  editor_set_tempo(player->editor);
      }
    }

    /* Check whether this thread should be killed */
    g_mutex_lock(player->mutex);
    if (player->killthread)
      break;
    g_mutex_unlock(player->mutex);
  }

  if (player->sched == SCHED_RTC)
    ioctl(player->rtc, RTC_PIE_OFF, 0);

  /* This is here because sometimes the player didn't stop all notes. Reason:
   * unknown. Seemed like a flush bug somewhere. */
  usleep(25000);

  /* Stop all playing notes */
  player_stop_notes(player);

  /* This is here because sometimes the player didn't stop all notes. Reason:
   * unknown. Seemed like a flush bug somewhere. */
  usleep(25000);

  /* The mutex is locked if the thread was killed and loop broken */
  g_mutex_unlock(player->mutex);

  /* Tell the editor playing has stopped */
  editor_stop(player->editor);
}

/* Creates a player for a song */
struct player *player_open(struct song *song, struct editor *editor)
{
  struct player *player = (struct player *)calloc(1, sizeof(struct player));
  int i;

  player->song = song;
  player->editor = editor;
  player->midi = midi_open(gconf_client_get_string(editor->gconfclient, "/apps/tutka/mididevice", NULL));
  player->sched = gconf_client_get_int(editor->gconfclient, "/apps/tutka/scheduler", NULL);
  if(player->sched == SCHED_RTC) {
    player->rtc = open("/dev/rtc", O_RDONLY);
    if (player->rtc != -1)
      player->sched = SCHED_RTC;
    else
      player->sched = SCHED_NANOSLEEP;
  } else
    player->rtc = -1;
  player->mutex = g_mutex_new();

  /* Initialize notes playing array */
  player_trackstatus_create(player, 0);

  /* Send messages to be autosent */
  for (i = 0; i < player->song->nummessages; i++) {
    if (player->song->messages[i]->autosend)
      midi_write(player->midi, player->song->messages[i]->data, player->song->messages[i]->length);
  }

  /* Reset MIDI controller values */
  for (i = 0; i < 16 * VALUES; i++)
    player->midicontrollervalues[i] = 0;

  return player;
}

/* Closes a player for a song */
void player_close(struct player *player)
{
  if (player == NULL) {
    fprintf(stderr, "player_close() called with null player\n");
    return;
  }

  /* Stop the player */
  player_stop(player);

  /* Free the mutex */
  if (player->mutex != NULL)
    g_mutex_free(player->mutex);

  /* Free the track status array */
  if (player->trackstatus != NULL)
    free(player->trackstatus);

  /* Close RTC */
  if (player->rtc != -1)
    close(player->rtc);

  /* Close MIDI */
  if (player->midi != NULL)
    midi_close(player->midi);

  free(player);
}

/* Starts the player thread */
void player_start(struct player *player, unsigned int mode, int section, int position, int block, int cont)
{
  if (player == NULL) {
    fprintf(stderr, "player_start() called with null player\n");
    return;
  }

  player_stop(player);

  player->mode = mode;
  player->tick = 0;

  switch (mode) {
  case MODE_PLAY_SONG:
    if (cont) {
      player->section = section;
      player->position = position;
    } else {
      player->section = 0;
      player->position = 0;
      player->line = 0;
    }
    player_refresh_playseq_and_block(player);
    break;
  case MODE_PLAY_BLOCK:
    player->block = block;
    if (!cont)
      player->line = 0;
    break;
  }

  /* Get the starting time */
  gettimeofday(&player->playingstarted, NULL);
  if (!cont) {
    player->playedsofar.tv_sec = 0;
    player->playedsofar.tv_usec = 0;
  }

  /* Create a new thread only if there isn't one already */
  if (player->thread == NULL) {
    player->thread = g_thread_create((GThreadFunc) player_thread, player, TRUE, NULL);
    g_thread_set_priority(player->thread, G_THREAD_PRIORITY_URGENT);
  }
}

/* Kills the player thread */
void player_stop(struct player *player)
{
  if (player == NULL) {
    fprintf(stderr, "player_stop() called with null player\n");
    return;
  }

  if (player->mode != MODE_IDLE)
    player->mode = MODE_IDLE;

  if (player->thread != NULL) {
    g_mutex_lock(player->mutex);
    player->killthread = 1;
    g_mutex_unlock(player->mutex);
    g_thread_join(player->thread);
    player->thread = NULL;
    player->killthread = 0;
  } else
    player_stop_notes(player);
}

/* Reallocate track status array */
void player_trackstatus_create(struct player *player, int oldmax)
{
  int i;

  if (player == NULL) {
    fprintf(stderr, "player_trackstatus_create() called with null player\n");
    return;
  }

  player->trackstatus = (char *)realloc(player->trackstatus, TRACK_STATUS_NUMVALUES * player->song->maxtracks * sizeof(char));

  /* Set new tracks to -1 */
  for (i = oldmax; i < player->song->maxtracks; i++) {
    player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_MIDI_CHANNEL] = -1;
    player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_NOTE] = -1;
    player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_VOLUME] = -1;
    player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_HOLD] = -1;
    player->trackstatus[i * TRACK_STATUS_NUMVALUES + TRACK_STATUS_PREVIOUS_COMMAND] = 0;
  }
}

/* Set current section */
void player_set_section(struct player *player, int section)
{
  if (player == NULL) {
    fprintf(stderr, "player_set_section() called with null player\n");
    return;
  }

  /* Bounds checking */
  if (section < 0)
    section = 0;
  if (section >= player->song->numsections)
    section = player->song->numsections - 1;

  g_mutex_lock(player->mutex);
  player->section = section;
  g_mutex_unlock(player->mutex);
}

/* Set current playseq */
void player_set_playseq(struct player *player, int playseq)
{
  if (player == NULL) {
    fprintf(stderr, "player_set_playseq() called with null player\n");
    return;
  }

  /* Bounds checking */
  if (playseq < 0)
    playseq = 0;
  if (playseq >= player->song->numplayseqs)
    playseq = player->song->numplayseqs - 1;

  g_mutex_lock(player->mutex);
  player->playseq = playseq;
  g_mutex_unlock(player->mutex);
}

/* Set current position */
void player_set_position(struct player *player, int position)
{
  if (player == NULL) {
    fprintf(stderr, "player_set_position() called with null player\n");
    return;
  }

  /* Bounds checking */
  if (position < 0)
    position = 0;
  if (position >= player->song->playseqs[player->playseq]->length)
    position = player->song->playseqs[player->playseq]->length - 1;

  g_mutex_lock(player->mutex);
  player->position = position;
  g_mutex_unlock(player->mutex);
}

/* Set current block */
void player_set_block(struct player *player, int block)
{
  if (player == NULL) {
    fprintf(stderr, "player_set_block() called with null player\n");
    return;
  }

  /* Bounds checking */
  if (block < 0)
    block = 0;
  if (block >= player->song->numblocks)
    block = player->song->numblocks - 1;

  g_mutex_lock(player->mutex);
  player->block = block;
  g_mutex_unlock(player->mutex);
}

/* Set current line */
void player_set_line(struct player *player, int line)
{
  if (player == NULL) {
    fprintf(stderr, "player_set_line() called with null player\n");
    return;
  }

  /* Bounds checking */
  while (line < 0)
    line += player->song->blocks[player->block]->length;
  while (line >= player->song->blocks[player->block]->length)
    line -= player->song->blocks[player->block]->length;

  g_mutex_lock(player->mutex);
  player->line = line;
  g_mutex_unlock(player->mutex);
}
